/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it
 * is applied to this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * AgroSense. If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.sensor.cropcircle;

import au.com.bytecode.opencsv.bean.ColumnPositionMappingStrategy;
import au.com.bytecode.opencsv.bean.CsvToBean;
import com.vividsolutions.jts.geom.Coordinate;
import eu.agrosense.client.grid.Grid;
import eu.agrosense.client.grid.GridProvider;
import java.beans.PropertyEditor;
import java.beans.PropertyEditorManager;
import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.MissingResourceException;
import java.util.logging.Logger;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import nl.cloudfarming.client.imp.ImportProcessor;
import nl.cloudfarming.client.logging.AppLogFactory;
import nl.cloudfarming.client.logging.AppLogger;
import nl.cloudfarming.client.model.EnterpriseIdProvider;
import nl.cloudfarming.client.model.SensorMetadata;
import nl.cloudfarming.client.model.SensorObservation;
import nl.cloudfarming.client.sensor.AbstractSensorImportProcessor;
import nl.cloudfarming.client.sensor.SensorDataObject;
import nl.cloudfarming.client.sensor.SensorPalette;
import org.netbeans.api.progress.ProgressHandle;
import org.netbeans.api.progress.ProgressHandleFactory;
import org.netbeans.api.project.Project;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.util.ImageUtilities;
import org.openide.util.Lookup;
import org.openide.util.NbBundle;
import org.openide.util.lookup.ServiceProvider;

/**
 * Processor for cropcircle data
 *
 * @author Timon Veenstra
 */
@ServiceProvider(service = ImportProcessor.class)
@NbBundle.Messages({"sensor cropcircle import error=Failed to import sensor data from file {0}", "sensor import format description=Cropcircle Sensor file (*.csv)"})
public class CropcircleImportProcessor extends AbstractSensorImportProcessor {

    public static final String CROPCIRCLE_FOLDER = "Cropcircle";
    private static final Logger LOGGER = Logger.getLogger(CropcircleImportProcessor.class.getName());
    private static final ImageIcon ICON = ImageUtilities.loadImageIcon(CropCircleDataObject.ICON_BASE, true);
    // Messages
    private static final String LOGKEY_FAILED_TO_READ_FILE = NbBundle.getMessage(SensorDataObject.class, "sensor.common.fileprocessor.logerror.failed_to_read_file");
    private static final String LOGKEY_PROCESSING_SUMMARY_ERRORMSG = NbBundle.getMessage(SensorDataObject.class, "sensor.common.fileprocessor.logerror.nr_of_records_failed");
    private static final String LOGKEY_PROCESSING_START = NbBundle.getMessage(SensorDataObject.class, "sensor.common.fileprocessor.loginfo.process_start");
    private static final String LOGKEY_PROCESSING_SUMMARY_PROCESSEDMSG = NbBundle.getMessage(SensorDataObject.class, "sensor.common.fileprocessor.loginfo.nr_of_processed");
    // Other variable declarations
    private static final AppLogger LOG = AppLogFactory.getLogger(CropcircleImportProcessor.class);
    PropertyEditor old;
    private static final String LOGKEY_TYPE_DISPLAYNAME = NbBundle.getMessage(CropcircleImportProcessor.class, "sensor.cropcircle.fileprocessor.asc340.log.displayname");
    private static final String[] COLUMS = new String[]{"longitude", "latitude", "fixType", "utcTime", "speed", "course", "sensor", "vi1", "vi2", "r1", "r2", "r3", "appRate"};
    private CsvToBean csv;
    private ColumnPositionMappingStrategy strat;

    public CropcircleImportProcessor() {

        old = PropertyEditorManager.findEditor(Date.class);

        strat = new ColumnPositionMappingStrategy();
        strat.setType(CropcircleSensorObservation.class);
        strat.setColumnMapping(COLUMS);
        csv = new CsvToBean();
    }

    @Override
    public String[] getSupportedExtentions() {
        return new String[]{"csv", "txt"};
    }

    @Override
    public Icon getIcon() {
        return ICON;
    }

    /**
     * Supports extensions csv and txt
     *
     * @param sensorFile
     * @return
     */
    @Override
    public boolean isSupported(FileObject sensorFile) {
        String MIMEType = FileUtil.getMIMEType(sensorFile);
        if (!CropCircleDataObject.MIME_TYPE.equals(MIMEType)) {
            return false;
        }
        return true;
    }

    @Override
    public String getFormatDescription() {
        return Bundle.sensor_import_format_description();
    }

    @Override
    public void process(FileObject importFile, Project project) throws IllegalArgumentException {
        process(importFile, project, new ArrayList());
    }

    @Override
    public void process(FileObject importFile, Project project, List toPopulate) throws IllegalArgumentException {

        //set temporary editor for date
        PropertyEditorManager.registerEditor(Date.class, UtcTimePropertyEditor.class);
        List<SensorObservation> sensorData = new ArrayList<>();

        try {
            int success = 0;
            int progress = 0;
            int fail = 0;

            LOG.info(LOGKEY_PROCESSING_START, getLogDisplayName(), importFile.getName());

            ProgressHandle progressHandle = ProgressHandleFactory.createHandle(NbBundle.getMessage(SensorDataObject.class, "sensor.common.fileprocessor.loginfo.process_description", getFormatDescription()));
            try {

                progressHandle.start(Long.valueOf(importFile.asLines().size()).intValue());

                for (String line : importFile.asLines()) {
                    progressHandle.progress(++progress);
                    if (!line.equals(getHeader())) {
                        try {
                            sensorData.add(parse(line));
                            success++;
                        } catch (Exception ex) {
                            fail++;
                        }
                    }
                }
                LOG.info(LOGKEY_PROCESSING_SUMMARY_PROCESSEDMSG, success);
                if (fail > 0) {
                    LOG.warning(LOGKEY_PROCESSING_SUMMARY_ERRORMSG, fail);
                }
            } finally {
                progressHandle.finish();
            }
        } catch (MissingResourceException | IOException ex) {
            LOG.error(LOGKEY_FAILED_TO_READ_FILE, importFile.getName());
            LOG.trace(ex);
        } finally {
            //revert back to default date property editor
            PropertyEditorManager.registerEditor(Date.class, old != null ? old.getClass() : null);
        }
        if (!sensorData.isEmpty()) {
            double minx = sensorData.get(0).getLongitude();
            double maxx = sensorData.get(0).getLongitude();
            double miny = sensorData.get(0).getLatitude();
            double maxy = sensorData.get(0).getLatitude();

            for (SensorObservation sd : sensorData) {
                minx = (minx < sd.getLongitude()) ? minx : sd.getLongitude();
                miny = (miny < sd.getLatitude()) ? miny : sd.getLatitude();
                maxx = (maxx > sd.getLongitude()) ? maxx : sd.getLongitude();
                maxy = (maxy > sd.getLatitude()) ? maxy : sd.getLatitude();
            }

            GridProvider gridProvider = Lookup.getDefault().lookup(GridProvider.class);
            Grid grid = gridProvider.getDefaultGridSection(project.getLookup().lookup(EnterpriseIdProvider.class), new Coordinate(minx, miny), new Coordinate(maxx, maxy));

            for (SensorObservation sd : sensorData) {
                //TODO sample data gets casted into an int, so multiply with 1000 for example
                if (sd.getNdvi() != null) {
                    grid.setRasterValue(new Coordinate(sd.getLongitude(), sd.getLatitude()), (int) (sd.getNdvi() * SensorPalette.SENSOR_DATA_MULTIPLIER));
                }

                if (sd.getWdvi() != null) {
                    grid.setRasterValue(new Coordinate(sd.getLongitude(), sd.getLatitude()), (int) (sd.getWdvi() * SensorPalette.SENSOR_DATA_MULTIPLIER));
                }
            }
            toPopulate.add(grid);
        }
    }

    /**
     * Parses payload to CropcircleSensorObservation
     *
     * @param payload Cropcircle ASC340 sensor line
     * @return CropcircleSensorObservation object
     * @throws CropcircleParseException
     */
    public CropcircleSensorObservation parse(String payload) throws CropcircleParseException {
        try {
            List list = csv.parse(strat, new StringReader(payload));
            return (CropcircleSensorObservation) list.get(0);
        } catch (RuntimeException ex) {
            throw new CropcircleParseException(ex.getCause());
        }
    }

    /**
     * Concrete subclasses must override this method to return the displayname
     * of that sensorprocessor to be displayed in the log. For example: "ASC-340
     * Cropcircle"
     *
     * Primary use is for logging, for example: [Info] Start processing of
     * ASC-340 Cropcircle sensor file 010710AD.CSV
     *
     * @return The log display name
     */
    protected String getLogDisplayName() {
        return LOGKEY_TYPE_DISPLAYNAME;
    }

    /**
     * Get the header of the Cropcircle file
     *
     * @return a string containing the expected header
     */
    public String getHeader() {
        StringBuilder header = new StringBuilder();
        CropcircleASC340FileDescription[] values = CropcircleASC340FileDescription.values();
        for (CropcircleASC340FileDescription desription : values) {
            header.append(desription);
            header.append(",");
        }
        header.deleteCharAt(header.lastIndexOf(","));
        return header.toString();
    }

    /**
     * Method which will create a SensorMetadata object from the preferences:
     * CropcirclePreference.SENSOR_DATA_FILE_FORMAT
     * CropcirclePreference.CROPCIRCLE_SENSOR_HEIGTH
     * CropcirclePreference.NUMBER_OF_CROPCIRCLE_SENSORS
     * CropcirclePreference.CROPCIRCLE_SENSOR_SPREAD
     * CropcirclePreference.SENSOR_TYPE CropcirclePreference.SPECTRAL_CHANNELS
     *
     * @return created SensorMetadata object
     */
    protected SensorMetadata buildMetadataFromPreferences() {
        String fileFormat = getFileDateFormat();
        int sensorHeigth = Integer.parseInt(CropcirclePreference.CROPCIRCLE_SENSOR_HEIGTH.getValue());
        int numberOfSensors = Integer.parseInt(CropcirclePreference.NUMBER_OF_CROPCIRCLE_SENSORS.getValue());
        String cropCircleSensorSpread = CropcirclePreference.CROPCIRCLE_SENSOR_SPREAD.getValue();
        String sensorType = CropcirclePreference.SENSOR_TYPE.getValue();
        String spectralChannels = CropcirclePreference.SPECTRAL_CHANNELS.getValue();
        return new SensorMetadata(fileFormat, sensorHeigth, numberOfSensors, cropCircleSensorSpread, sensorType, spectralChannels);
    }

    protected String getFileDateFormat() {
        return CropcirclePreference.SENSOR_DATA_FILE_FORMAT.getValue();
    }

    @Override
    protected FileObject getSensorImportSubFolder(Project project) throws IOException {
        return getSubFolder(getSensorImportRootFolder(project), CROPCIRCLE_FOLDER);
    }
}
